# -*- coding: utf-8 -*-

from rest_framework.exceptions import APIException


class InvalidMinkeSetup(Exception):
    pass


class InvalidURLQuery(APIException):
    status_code = 400
    default_detail = 'Invalid url-query.'
    default_code = 'invalid_urlquery'



# -*- coding: utf-8 -*-

import json
from django.db import models
from django.core.serializers.json import DjangoJSONEncoder


def item_by_attr(list, attr, value, default=None):
    return next((i for i in list if hasattr(i, attr) and getattr(i, attr) == value), default)


def prepare_shell_command(cmd):
    # linux-shells need \n as newline
    return cmd.replace('\r\n', '\n').replace('\r', '\n')


class JSONField(models.TextField):
    """
    A very raw and simple JSONField.
    (It is only used internally - so we know what we get.)
    """
    def from_db_value(self, value, *args):
        if value is None:
            return value
        else:
            return json.loads(value)

    def to_python(self, value):
        if type(value) is str:
            return json.loads(value)
        else:
            return value

    def get_prep_value(self, value):
        return json.dumps(value, cls=DjangoJSONEncoder)


def get_session_summary(sessions):
    """
    Takes a list of sessions and extract a summary-dictonary.
    """
    summary = dict()
    summary['all'] = len(sessions)
    summary['waiting'] = len([s for s in sessions if s.proc_status == 'initialized'])
    summary['running'] = len([s for s in sessions if s.proc_status in ('running', 'stopping')])
    summary['done'] = len([s for s in sessions if s.proc_status in ('completed', 'stopped', 'canceled', 'failed')])
    summary['success'] = len([s for s in sessions if s.session_status == 'success'])
    summary['warning'] = len([s for s in sessions if s.session_status == 'warning'])
    summary['error'] = len([s for s in sessions if s.session_status == 'error'])
    return summary



# -*- coding: utf-8 -*-

import re
from collections import OrderedDict
from fabric2.runners import Result

from django.db.utils import OperationalError
from django.db.utils import ProgrammingError
from django.core.exceptions import FieldDoesNotExist
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.utils.text import camel_case_to_spaces

from .models import Host
from .models import MinkeModel
from .models import MinkeSession
from .models import BaseMessage
from .models import CommandResult
from .forms import CommandForm
from .messages import PreMessage
from .messages import TableMessage
from .messages import ExecutionMessage
from .exceptions import InvalidMinkeSetup


class REGISTRY(OrderedDict):
    """
    A reload-able session-registry.
    """
    _session_factories = list()

    @classmethod
    def add_session_factory(cls, factory):
        """
        Add a session-factory (any callable that declares session-classes).
        """
        cls._session_factories.append(factory)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._dict = None

    def reload(self):
        """
        Reload the registry by running the factories.
        """
        # Before running the factories we restore the original registry.
        # So we are up-to-date without messing up.
        if self._dict:
            self.clear()
            self.update(self._dict)
        else:
            self._dict = self.copy()
        for factory in self._session_factories:
            factory()


REGISTRY = REGISTRY()


class SessionRegistration(type):
    """
    metaclass for Sessions that implements session-registration
    """
    def __new__(cls, name, bases, dct):
        # Setting the abstract-attr explicitly avoids its inheritance.
        dct['abstract'] = dct.get('abstract', False)
        return super().__new__(cls, name, bases, dct)

    def __init__(cls, classname, bases, attrs):
        super().__init__(classname, bases, attrs)
        if not cls.abstract and not cls.__name__ in REGISTRY:
            cls.register()
            if cls.add_permission: cls.create_permission()

    def register(cls):
        """
        Register the session-class.
        """
        # some sanity-checks
        if not cls.work_on:
            msg = 'At least one minke-model must be specified for a session.'
            raise InvalidMinkeSetup(msg)

        for model in cls.work_on:
            try:
                assert(model == Host or issubclass(model, MinkeModel))
            except (TypeError, AssertionError):
                msg = '{} is no minke-model.'.format(model)
                raise InvalidMinkeSetup(msg)

        if issubclass(cls, SingleCommandSession) and not cls.command:
            msg = 'SingleCommandSession needs to specify an command.'
            raise InvalidMinkeSetup(msg)

        if issubclass(cls, CommandChainSession) and not cls.commands:
            msg = 'CommandChainSession needs to specify commands.'
            raise InvalidMinkeSetup(msg)

        if issubclass(cls, SessionChain) and not cls.sessions:
            msg = 'SessionChain needs to specify sessions.'
            raise InvalidMinkeSetup(msg)

        # set verbose-name if missing
        if not cls.verbose_name:
            cls.verbose_name = camel_case_to_spaces(cls.__name__)

        # register session
        REGISTRY[cls.__name__] = cls

    def _get_permission(cls):
        codename = 'run_{}'.format(cls.__name__.lower())
        name = 'Can run {}'.format(camel_case_to_spaces(cls.__name__))
        lookup = 'minke.{}'.format(codename)
        return codename, name, lookup

    def create_permission(cls):
        """
        Create a run-permission for this session-class.
        """
        # create session-permission...
        # In some contexts get_for_model fails because content_types aren't
        # setup. This happens when applying migrations but also when testing
        # with sqlite3.
        try: content_type = ContentType.objects.get_for_model(MinkeSession)
        except (OperationalError, ProgrammingError): return
        codename, name, lookup = cls._get_permission()
        permission = Permission.objects.get_or_create(
            codename=codename,
            content_type=content_type,
            defaults=dict(name=name))
        cls.permissions += (lookup,)

    def delete_permission(cls):
        codename, name, lookup = cls._get_permission()
        try:
            Permission.objects.get(codename=codename).delete()
        except Permission.DoesNotExist:
            pass
        else:
            cls.permissions = tuple(set(cls.permissions) - set((lookup,)))



def protect(method):
    """
    Decorator for session-methods to protect them from being interrupted by a
    soft-interruption.
    """
    def wrapper(obj, *args, **kwargs):
        # are we already protected?
        if obj._busy: return method(obj, *arg, **kwargs)
        # otherwise protect the method-call by setting the busy-flag
        obj._busy = True
        result = method(obj, *args, **kwargs)
        # if interruption was deferred now is the time to raise it
        if obj._interrupt: raise obj._interrupt
        obj._busy = False
        return result

    return wrapper


class Session(metaclass=SessionRegistration):
    abstract = True
    verbose_name = None
    work_on = tuple()
    permissions = tuple()
    form = None
    confirm = False
    wait_for_execution = False
    add_permission = True
    invoke_config = dict()
    soft_interruption = False
    parrallel_per_host = False

    @classmethod
    def get_form(cls):
        return cls.form

    def __init__(self, con, db):
        self._con = con
        self._db = db
        self._interrupt = None
        self._busy = False
        self.start = db.start
        self.end = db.end

    def cancel(self, interrupt):
        """
        Either raise interrupt instantly, or - in case of a soft-interruption -
        let execute raise it when its work is done.
        """
        # NOTE: It seems that there is no chance to interrupt a shell-process
        # started by fabric if no pty is in use.
        # fabric.runners.Remote.send_interrupt says:
        # > ... in v1, we just reraised the KeyboardInterrupt unless a PTY was
        # > present; this seems to have been because without a PTY, the
        # > below escape sequence is ignored, so all we can do is immediately
        # > terminate on our end...
        # Therefore it makes most sense to use a soft interruption if the
        # run.pty-config is False. Otherwise we would instantly quit though
        # but the shel-proc would run on and we lose its response.
        self._interrupt = interrupt
        if not self.soft_interruption or not self._busy:
            raise self._interrupt

    @property
    def status(self):
        return self._db.session_status

    @property
    def minkeobj(self):
        return self._db.minkeobj

    @property
    def data(self):
        return self._db.session_data

    def process(self):
        """
        Real work is done here...
        """
        raise NotImplementedError('Your session must define a process-method!')

    def add_msg(self, msg, level=None):
        """
        Add a message to the session.
        You could either pass an instance of a message-class, or any type the
        message-classes are initiated with (string, tuple or result-object).
        """
        if isinstance(msg, str): msg = PreMessage(msg, level)
        elif isinstance(msg, tuple): msg = TableMessage(msg, level)
        elif isinstance(msg, Result): msg = ExecutionMessage(msg, level)
        elif isinstance(msg, BaseMessage): pass
        self._db.messages.add(msg, bulk=False)

    def set_status(self, status, alert=True):
        """
        Set session-status. Pass a valid session-status or a boolean.
        """
        states = dict(self._db.SESSION_STATES)
        if type(status) == bool:
            status = 'success' if status else 'error'
        elif status.lower() in states.keys():
            status = status.lower()
        else:
            msg = 'session-status must be one of {}'.format(states.keys())
            raise InvalidMinkeSetup(msg)

        if not self.status or not alert or states[self.status] < states[status]:
            self._db.session_status = status

    def format_cmd(self, cmd):
        """
        Will format a given command-string using the minkeobj's attributes
        and the session_data while the session_data has precedence.
        """
        params = vars(self.minkeobj)
        params.update(self.data)
        return cmd.format(**params)

    def valid(self, result, regex=None):
        """
        Validate result.

        Return True if rtn-code is 0.
        If regex is given it must also match stdout.
        """
        if regex and result.ok:
            return bool(re.match(regex, result.stdout))
        else:
            return result.ok

    def _run(self, cmd, **kwargs):
        """
        Run a command and save the result.
        """
        result = self._con.run(cmd, **kwargs)
        self._db.commands.add(result, bulk=False)
        return result

    @protect
    def run(self, cmd, **kwargs):
        """
        Run a command and return its response.
        """
        return self._run(cmd, **kwargs)

    @protect
    def execute(self, cmd, add_msg=True, set_status=True, **kwargs):
        """
        Run a command, leave a ExecutionMessage and set the session-status.
        """
        result = self._run(cmd, **kwargs)

        # choose session-status depending on result-characteristics
        if result.failed: status = 'error'
        elif result.stderr: status = 'warning'
        else: status = 'success'

        if add_msg: self.add_msg(result)
        if set_status: self.set_status(status)

        return result.ok


class UpdateEntriesSession(Session):
    abstract = True

    def update_field(self, field, cmd, regex=None):
        """
        Update a field using either the stdout or the first matching-group
        from regex.
        """
        # is field a minkeobj-attribute?
        try: getattr(self.minkeobj, field)
        except AttributeError as e: raise e

        # run cmd
        result = self.run(cmd)
        valid = self.valid(result, regex)

        # do we have a regex? Then use the first captured group if there is one,
        # stdout otherwise...
        if valid and regex and result.stdout:
            groups = re.match(regex, result.stdout).groups()
            value = groups[0] if groups else result.stdout

        # no regex? just take stdout as it is...
        elif valid and result.stdout:
            value = result.stdout

        # valid but no stdout? Leave a warning...
        elif valid and not result.stdout:
            self.add_msg(result, 'warning')
            value = None

        # not valid - error-message...
        else:
            self.add_msg(result, 'error')
            value = None

        setattr(self.minkeobj, field, value)
        return bool(value)


class SingleCommandSession(Session):
    abstract = True
    command = None

    def process(self):
        self.execute(self.format_cmd(self.command))


class CommandFormSession(SingleCommandSession):
    abstract = True
    form = CommandForm
    command = '{cmd}'


class CommandChainSession(Session):
    abstract = True
    commands = tuple()
    break_states = ('error',)

    def process(self):
        for cmd in self.commands:
            self.execute(self.format_cmd(cmd))
            if self.status in self.break_states:
                break


class SessionChain(Session):
    abstract = True
    sessions = tuple()
    break_states = ('error',)

    def process(self):
        for cls in self.sessions:
            session = cls(self._con, self._db)
            session.process()
            if session.status in self.break_states:
                break



# -*- coding: utf-8 -*-

from django.core.exceptions import FieldError
from django.template.loader import render_to_string

from rest_framework.response import Response
from rest_framework.generics import ListAPIView
from rest_framework.filters import BaseFilterBackend
from rest_framework.permissions import IsAuthenticated

from .serializers import SessionSerializer
from .exceptions import InvalidURLQuery
from .models import MinkeSession
from .utils import get_session_summary


class LookupFilter(BaseFilterBackend):
    """
    Use the url-query as lookup-params.
    """
    def get_lookup_params(self, request):
        params = dict()
        for k, v in request.GET.items():
            if k == 'summary': continue
            elif k.endswith('__in'): params[k] = v.split(',')
            else: params[k] = v
        return params

    def filter_queryset(self, request, queryset, view):
        lookup_params = self.get_lookup_params(request)
        try:
            return queryset.filter(**lookup_params)
        except (FieldError, ValueError):
            msg = 'Invalid lookup-parameter: {}'.format(lookup_params)
            raise InvalidURLQuery(msg)


class UserFilter(BaseFilterBackend):
    """
    Filter sessions by user.
    """
    def filter_queryset(self, request, queryset, view):
        return queryset.filter(user=request.user)


class SessionListAPI(ListAPIView):
    """
    API endpoint to retrieve sessions.
    """
    permission_classes = (IsAuthenticated,)
    serializer_class = SessionSerializer
    filter_backends = (LookupFilter, UserFilter)
    queryset = MinkeSession.objects.prefetch_related('messages')

    def list(self, request, *arg, **kwargs):
        """
        Either return json-formatted sessions or a html-summary-snippet.
        """
        if 'summary' in request.GET:
            sessions = list(self.filter_queryset(self.get_queryset()))
            summary = get_session_summary(sessions)
            context = dict(session_count=summary)
            summary_html = render_to_string('minke/session_summary.html', context)
            return Response(summary_html)
        else:
            return super().list(request, *arg, **kwargs)

    def put(self, request, *arg, **kwargs):
        """
        The put-apicall is used to cancel initialized or running sessions.
        """
        queryset = self.filter_queryset(self.get_queryset())

        canceled_sessions = list()
        for session in queryset:
            if session.proc_status in ['initialized', 'running']:
                canceled = session.cancel()
                if canceled: canceled_sessions.append(session)

        serializer = self.get_serializer(canceled_sessions, many=True)
        return Response(serializer.data)



# -*- coding: utf-8 -*-

from fabric2.config import Config
from fabric2.runners import Remote

from .exceptions import InvalidMinkeSetup
from .models import CommandResult


class FabricConfig(Config):
    """
    A subclass of fabric's Config-class.
    Add a load_snakeconfig-method, that takes a plain dict and parses its
    snake-case-keys to fit into the nested default-config-structure.
    """
    def load_snakeconfig(self, configdict):
        for param, value in configdict.items():
            # We support two level of recursive config-keys. That means
            # something like 'connect_kwargs_my_special_key' will be applied as
            # 'config.connect_kwargs.my_special_key'. The key on the first level
            # must already exist. Otherwise we raise InvalidMinkeSetup.
            snippets = param.split('_')
            key1 = key2 = None

            for i, key in enumerate(snippets):
                if '_'.join(snippets[:i+1]) in self:
                    key1 = '_'.join(snippets[:i+1])
                    key2 = '_'.join(snippets[i+1:])
                    break

            if not key1:
                msg = 'Invalid fabric-config-parameter: {}'.format(param)
                raise InvalidMinkeSetup(msg)

            # apply config-data
            if not key2:
                self[key1] = value
            else:
                if not self[key1]: self[key1] = dict()
                self[key1][key2] = value


class FabricRemote(Remote):
    """
    A subclass of fabric's remote-runner to customize the result-class.
    """
    def generate_result(self, **kwargs):
        kwargs["connection"] = self.context
        return CommandResult(**kwargs)



# -*- coding: utf-8 -*-

from django.db import models
from django.db.models import Prefetch
from django.contrib.contenttypes.models import ContentType
from django.contrib.auth.models import Permission
from django.db.models.signals import pre_delete
from django.db.models.signals import post_save
from django.dispatch import receiver

from minke.models import MinkeModel
from minke.models import MinkeSession
from minke.models import Host
from minke.utils import prepare_shell_command

from .sessions import BaseCommandSession
from .sessions import BaseCommandChoiceSession
from .sessions import BaseCommandChainSession


def get_minketypes():
    minketype_ids = list()
    excludes = ('auth', 'contenttypes', 'sessions', 'commands')
    for ct in ContentType.objects.exclude(app_label__in=excludes):
        if issubclass(ct.model_class(), MinkeModel) or ct.model_class() is Host:
            minketype_ids.append(ct.id)
    return dict(id__in=minketype_ids)


class CommandManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().prefetch_related('minketypes')


class CommandGroupManager(CommandManager):
    def get_queryset(self):
        cmd_qs = Command.objects.order_by('commandorder__order')
        prefetch = Prefetch('commands', queryset=cmd_qs)
        return super().get_queryset().prefetch_related(prefetch)


class BaseCommands(models.Model):
    class Meta:
        abstract = True

    label = models.CharField(max_length=256)
    description = models.TextField(blank=True)
    minketypes = models.ManyToManyField(ContentType, limit_choices_to=get_minketypes)
    active = models.BooleanField(default=True)

    def _get_session_attrs(self):
        attrs = dict()
        attrs['abstract'] = True
        attrs['model'] = self.__class__
        attrs['model_id'] = self.id
        attrs['verbose_name'] = self.label
        attrs['__doc__'] = self.description
        attrs['work_on'] = tuple((ct.model_class() for ct in self.minketypes.all()))
        return attrs

    def _get_session_class(self):
        return BaseCommandSession

    def _get_session_name(self):
        return '%s_%d' % (self.__class__.__name__, self.id)

    def as_session(self):
        name = self._get_session_name()
        attrs = self._get_session_attrs()
        baseclass = self._get_session_class()
        return type(name, (baseclass,), attrs)


class Command(BaseCommands):
    objects = CommandManager()
    cmd = models.TextField()

    def save(self, *args, **kwargs):
        self.cmd = prepare_shell_command(self.cmd)
        return super().save(*args, **kwargs)

    def __str__(self):
        return self.label

    def _get_session_attrs(self):
        attrs = super()._get_session_attrs()
        attrs['command'] = self.cmd
        return attrs


class CommandGroup(BaseCommands):
    objects = CommandGroupManager()
    commands = models.ManyToManyField(Command, through='CommandOrder')
    as_options = models.BooleanField(default=False)

    def __str__(self):
        return self.label

    def _get_session_attrs(self):
        attrs = super()._get_session_attrs()
        if not self.as_options:
            attrs['commands'] = tuple(self.commands.all())
        return attrs

    def _get_session_class(self):
        if self.as_options:
            return BaseCommandChoiceSession
        else:
            return BaseCommandChainSession


@receiver(pre_delete)
def delete_permission(sender, instance, **kwargs):
    """
    Delete the run-session-permission.
    """
    if not sender in (Command, CommandGroup): return
    session = instance.as_session()
    session.delete_permission()


@receiver(post_save)
def create_permission(sender, instance, **kwargs):
    """
    Create the run-session-permission.
    """
    if not sender in (Command, CommandGroup): return
    session = instance.as_session()
    session.create_permission()


class CommandOrder(models.Model):
    command = models.ForeignKey(Command, on_delete=models.CASCADE)
    commands = models.ForeignKey(CommandGroup, on_delete=models.CASCADE)
    order = models.SmallIntegerField()

    class Meta:
        unique_together = ('commands', 'order')
        get_latest_by = 'order'

    def get_order(self):
        try:
            latest = CommandOrder.objects.filter(commands=self.commands).latest()
        except CommandOrder.DoesNotExist:
            return 1
        else:
            return latest.order + 1

    def save(self, *args, **kwargs):
        self.order = self.get_order()
        super().save(*args, **kwargs)



# -*- coding: utf-8 -*-

from django.contrib.contenttypes.models import ContentType
from minke.sessions import CommandFormSession
from minke.sessions import SingleCommandSession
from minke.sessions import CommandChainSession
from minke.sessions import SessionRegistration
from minke.sessions import REGISTRY
from minke.models import Host
from .forms import CommandForm


class BaseCommandSession(SingleCommandSession):
    add_permission = False
    abstract = True
    model = None
    model_id = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        obj = self.model.objects.get(pk=self.model_id)
        self.command = obj.cmd


class BaseCommandChainSession(CommandChainSession):
    add_permission = False
    abstract = True
    model = None
    model_id = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        obj = self.model.objects.get(pk=self.model_id)
        self.commands = (c.cmd for c in obj.commands.all())


class BaseCommandChoiceSession(CommandFormSession):
    add_permission = False
    abstract = True
    model = None
    model_id = None
    form = CommandForm

    @classmethod
    def get_form(cls):
        obj = cls.model.objects.get(pk=cls.model_id)
        choices = ((c.id, str(c)) for c in obj.commands.all())
        cls.form.base_fields['cmd'].choices = choices
        return cls.form

    def format_cmd(self, cmd):
        cmd_obj = self.model.commands.get(pk=self.data['cmd'])
        return super().format_cmd(cmd_obj.cmd)


def session_factory():
    from .models import Command
    from .models import CommandGroup
    for obj in Command.objects.filter(active=True):
        session = obj.as_session()
        session.register()

    for obj in CommandGroup.objects.filter(active=True):
        session = obj.as_session()
        session.register()


REGISTRY.add_session_factory(session_factory)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.12 on 2018-06-19 17:03
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('minke', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='host',
            options={'ordering': ['host']},
        ),
    ]



# -*- coding: utf-8 -*-
from pydoc import locate

from django import forms
from django.contrib.admin.options import IncorrectLookupParameters
from django.contrib.contenttypes.models import ContentType
from django.contrib.admin.views.main import ChangeList
from django.contrib.admin.filters import RelatedOnlyFieldListFilter
from django.contrib.admin import helpers
from django.contrib import messages
from django.contrib import admin
from django.utils.text import Truncator
from django.utils.html import escape
from django.utils.translation import gettext as _
from django.utils.safestring import mark_safe
from django.template.response import TemplateResponse
from django.core.exceptions import PermissionDenied
from django.db.models import Prefetch
from django.urls import reverse
from django.http import HttpResponseRedirect

from . import settings
from . import engine
from .sessions import REGISTRY
from .models import MinkeSession
from .models import Host
from .models import HostGroup
from .models import CommandResult
from .models import BaseMessage
from .forms import MinkeForm
from .forms import SessionSelectForm
from .filters import StatusFilter
from .utils import get_session_summary


class SessionChangeList(ChangeList):
    """
    A changelist to support additional get-parameters.
    """
    def get_filters_params(self, params=None):
        params = super().get_filters_params(params)
        for key in ('display_messages', 'display_commands'):
            if key in params: del params[key]
        return params

    def get_queryset(self, request):
        queryset = super().get_queryset(request)
        return queryset.prefetch_related('messages', 'commands', 'minkeobj')


@admin.register(MinkeSession)
class SessionAdmin(admin.ModelAdmin):
    model = MinkeSession
    change_list_template = 'minke/change_list.html'

    def get_changelist(self, request, **kwargs):
        """
        Use the SessionChangeList to support additional get-parameters.
        """
        return SessionChangeList

    class Media:
        css = dict(all=('minke/css/minke.css',))

    date_hierarchy = 'created_time'
    search_fields = ('session_name',)
    list_display = (
        'session_verbose_name',
        'minkeobj_view',
        'minkeobj_type',
        'session_status',
        'proc_status',
        'start_time',
        'run_time')
    list_filter = (
        'session_name',
        'session_status',
        'proc_status',
        ('minkeobj_type', RelatedOnlyFieldListFilter),
        'user')
    fieldsets = (
        (None, {
            'fields': (
                'minkeobj_view',
                'minkeobj_type',
                'user',
            ),
            'classes': ('extrapretty', 'wide')
        }),
        (_('Session'), {
            'fields': (
                'session_name',
                'session_verbose_name',
                'session_status',
                'session_description',
                'session_data',
            ),
            'classes': ('extrapretty', 'wide')
        }),
        (_('Processing'), {
            'fields': (
                'proc_status',
                'task_id',
                'created_time',
                'start_time',
                'end_time',
                'run_time',
            ),
            'classes': ('extrapretty', 'wide')
        }),
    )
    readonly_fields = fieldsets[0][1]['fields'] + fieldsets[1][1]['fields'] + fieldsets[2][1]['fields']

    def minkeobj_view(self, obj):
        opts = obj.minkeobj._meta
        lookup = "admin:{}_{}_change".format(opts.app_label, opts.model_name)
        href = reverse(lookup, args=(obj.minkeobj.id,))
        return mark_safe('<a href="{}">{}</a>'.format(href, obj.minkeobj))
    minkeobj_view.short_description = 'Minke-object'
    minkeobj_view.admin_order_field = 'minkeobj_id'

    def changelist_view(self, request, extra_context=None):
        display_messages = bool(int(request.GET.get('display_messages', 0)))
        display_commands = bool(int(request.GET.get('display_commands', 0)))
        extra_context = extra_context or dict()
        extra_context['display_session_switch'] = True
        extra_context['display_session_info'] = display_messages or display_commands
        extra_context['display_messages'] = display_messages
        extra_context['display_commands'] = display_commands
        return super().changelist_view(request, extra_context)


class MinkeChangeList(ChangeList):
    """
    Subclass ChangeList to build the session-list based on the result_list.
    """
    def __init__(self, request, *args, **kwargs):
        super().__init__(request, *args, **kwargs)
        # We need a plain session-list with the same order as the result_list.
        # They will be zipped with the results coming from the result_list-templatetag.
        sessions = [(list(o.sessions.all())+[None])[0] for o in self.result_list]
        self.sessions = sessions
        self.session_count = get_session_summary([s for s in sessions if not s is None])

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        currents = MinkeSession.objects.filter(current=True, user=request.user)
        currents = currents.prefetch_related('messages', 'commands')
        return qs.prefetch_related(Prefetch('sessions', queryset=currents))


class MinkeAdmin(admin.ModelAdmin):
    """
    MinkeAdmin is the ModelAdmin-baseclass for MinkeModels.
    It allows to run sessions from the changelist-view.
    """
    change_form_template = 'minke/change_form.html'
    change_list_template = 'minke/change_list.html'
    session_select_form = SessionSelectForm

    class Media:
        css = dict(all=('minke/css/minke.css',))

    def get_changelist(self, request, **kwargs):
        """
        Use our own ChangeList.
        """
        return MinkeChangeList

    def get_changelist_instance(self, request):
        """
        Normalize the way to get a changelist-instance. Prior django-2.0 the
        get_changelist_instance-method is missing.
        """
        try:
            return super().get_changelist_instance(request)
        except AttributeError:
            # prior django-2.0 get_changelist_instance does not exists
            list_display = self.get_list_display(request)
            list_display_links = self.get_list_display_links(request, list_display)
            list_filter = self.get_list_filter(request)
            search_fields = self.get_search_fields(request)
            list_select_related = self.get_list_select_related(request)

            # Check actions to see if any are available on this changelist
            actions = self.get_actions(request)
            if actions:
                # Add the action checkboxes if there are any actions available.
                list_display = ['action_checkbox'] + list(list_display)

            ChangeList = self.get_changelist(request)
            return ChangeList(
                request, self.model, list_display,
                list_display_links, list_filter, self.date_hierarchy,
                search_fields, list_select_related, self.list_per_page,
                self.list_max_show_all, self.list_editable, self)

    def permit_session(self, request, session):
        permitted = request.user.has_perms(session.permissions)
        permitted &= self.model in session.work_on
        return permitted

    def get_session_options(self, request):
        """
        Get sessions  valid for the user and model of this request.
        Return a list of tuples like with session-name and -verbose-name.
        """
        REGISTRY.reload()
        sessions = [(None, '---------')]

        # filter sessions in respect to their permissions- and work_on-attrs
        for session in REGISTRY.values():
            if not self.permit_session(request, session): continue
            sessions.append((session.__name__, session.verbose_name))
        return sessions

    def get_session_select_form(self, request, data=None):
        """
        Return SessionSelectForm-instance with apropriate session-choices.
        """
        form = self.session_select_form(data or dict())
        form.fields['session'].choices = self.get_session_options(request)
        return form

    def get_session_cls(self, request):
        session_name = request.POST.get('session', None)
        if not session_name:
            msg = _("No session were selected that should be run.")
            self.message_user(request, msg, messages.WARNING)
            return None
        else:
            REGISTRY.reload()
            return REGISTRY.get(session_name, None)

    def run_sessions(self, request, session_cls, queryset, force_confirm=False):
        wait = session_cls.wait_for_execution
        confirm = force_confirm or session_cls.confirm
        fabric_config = None
        session_data = dict()
        session_form_cls = session_cls.get_form()
        fabric_form_cls = None
        render_params = dict()

        # import fabric-form if needed...
        if settings.MINKE_FABRIC_FORM:
            fabric_form_cls = locate(settings.MINKE_FABRIC_FORM)
            if not fabric_form_cls:
                msg = '{} could not be loaded'.format(settings.MINKE_FABRIC_FORM)
                raise InvalidMinkeSetup(msg)

        # Do we have to work with a minke-form?
        if confirm or fabric_form_cls or session_form_cls:

            # Do we come from a minke-form?
            from_form = request.POST.get('minke_form', False)
            if from_form:
                minke_form = MinkeForm(request.POST, auto_id=False)
                valid = minke_form.is_valid()
                form_data = [request.POST]
            else:
                minke_form = MinkeForm(initial=request.POST, auto_id=False)
                valid = False
                form_data = list()

            # initiate fabric-form
            if fabric_form_cls:
                fabric_form = fabric_form_cls(*form_data, auto_id=False)
                render_params['fabric_form'] = fabric_form
                valid &= fabric_form.is_valid()

            # initiate session-form
            if session_form_cls:
                session_form = session_cls.form(*form_data, auto_id=False)
                render_params['session_form'] = session_form
                valid &= session_form.is_valid()

            # render minke-form first time or if form-data is invalid
            if not from_form or not valid:
                render_params['title'] = session_cls.verbose_name,
                render_params['minke_form'] = minke_form
                render_params['objects'] = queryset
                render_params['object_list'] = confirm
                return TemplateResponse(request, 'minke/minke_form.html', render_params)

            else:
                # collect form-data
                if fabric_form_cls: fabric_config = fabric_form.cleaned_data
                if session_form_cls: session_data = session_form.cleaned_data

        # lets rock...
        engine.process(session_cls, queryset, session_data, request.user,
                       fabric_config=fabric_config, wait=wait)

    def changelist_view(self, request, extra_context=None):
        """
        Extend the modeladmin-changelist_view by session-processing.
        """
        extra_context = extra_context or dict()
        extra_context['display_session_select'] = extra_context.get('display_session_select', True)
        extra_context['display_session_info'] = extra_context.get('display_session_info', True)
        extra_context['display_session_proc_info'] = extra_context.get('display_session_proc_info', True)
        extra_context['display_messages'] = extra_context.get('display_messages', True)
        extra_context['display_commands'] = extra_context.get('display_commands', False)

        # Does this request has something to do with sessions at all?
        if ('run_sessions' not in request.POST
        and 'clear_sessions' not in request.POST):
            return super().changelist_view(request, extra_context)

        # setup
        force_confirm = False
        selected = request.POST.getlist(helpers.ACTION_CHECKBOX_NAME)
        select_across = bool(int(request.POST.get('select_across', 0)))
        session_form = self.get_session_select_form(request, request.POST)
        redirect_url = request.get_full_path()

        try:
            cl = self.get_changelist_instance(request)
        except IncorrectLookupParameters:
            # Since we are coming from a form-subimission our url-query
            # should be valid. So this shouldn't happen at all.
            return HttpResponseRedirect(redirect_url)

        # Do we have any selected items? Else leave a message and redirect.
        if not selected:
            msg = _("No items selected.")
            self.message_user(request, msg, messages.WARNING)
            return HttpResponseRedirect(redirect_url)

        # get queryset
        queryset = cl.get_queryset(request)
        if not select_across:
            queryset = queryset.filter(pk__in=selected)

        # clear session-infos for selected items
        if 'clear_sessions' in request.POST:
            MinkeSession.objects.clear_currents(request.user, queryset)
            return HttpResponseRedirect(redirect_url)

        # run sessions
        elif 'run_sessions' in request.POST:

            # get session_cls.
            session_form = self.get_session_select_form(request, request.POST)
            if session_form.is_valid():
                session_name = session_form.cleaned_data['session']
                REGISTRY.reload()
                session_cls = REGISTRY[session_name]
            else:
                msg = _("No session selected.")
                self.message_user(request, msg, messages.WARNING)
                return HttpResponseRedirect(redirect_url)

            # Do the user have permissions to run this session-type?
            if not self.permit_session(request, session_cls):
                raise PermissionDenied

            # If this is a select-across-request, we force confirmation
            # and redirect to a show-all-changelist.
            # FIXME: select-across should be limited at least to the value of
            # list_max_show_all.
            if select_across:
                force_confirm = True
                delimiter = '&' if '?' in redirect_url else '?'
                redirect_url += delimiter + 'all='

            # run_sessions might want to render a minke- or session-form
            response = self.run_sessions(request, session_cls, queryset, force_confirm)
            return response or HttpResponseRedirect(redirect_url)


def get_ssh_options():
    choices = ((None, '---------'),)
    choices += tuple(((k, k) for k in settings.MINKE_HOST_CONFIG.keys()))
    return choices


class HostAdminForm(forms.ModelForm):
    class Meta:
        model = Host
        exclude = tuple()
        widgets = dict(config=forms.Select(choices=get_ssh_options()))


class HostGroupAdminForm(forms.ModelForm):
    class Meta:
        model = HostGroup
        exclude = tuple()
        widgets = dict(config=forms.Select(choices=get_ssh_options()))


@admin.register(Host)
class HostAdmin(MinkeAdmin):
    model = Host
    form = HostAdminForm
    list_display = ('name', 'verbose_name', 'username', 'hostname', 'port', 'disabled')
    list_editable = ('disabled',)
    search_fields = ('name', 'hostname')
    ordering = ('name',)
    list_filter = (StatusFilter,)


@admin.register(HostGroup)
class HostGroupAdmin(admin.ModelAdmin):
    model = HostGroup
    form = HostGroupAdminForm
    list_display = ('name',)
    search_fields = ('name',)
    ordering = ('name',)


class CommandResultAdmin(admin.ModelAdmin):
    model = CommandResult
    list_display = ('command', 'session', 'exited', 'stdout_view', 'stderr_view', 'shell', 'pty', 'encoding', 'created_time')
    readonly_fields = list_display
    search_fields = ('session', 'command')
    ordering = ('session', 'created_time')

    def stdout_view(self, obj):
        if not obj.stdout: return
        abbr = escape(Truncator(obj.stdout).words(3))
        html = '<span title="{}">{}</span>'.format(escape(obj.stdout), abbr)
        return mark_safe(html)
    stdout_view.short_description = 'stdout'

    def stderr_view(self, obj):
        if not obj.stderr: return
        abbr = escape(Truncator(obj.stderr).words(3))
        html = '<span title="{}">{}</span>'.format(escape(obj.stderr), abbr)
        return mark_safe(html)
    stderr_view.short_description = 'stderr'


class BaseMessageAdmin(admin.ModelAdmin):
    model = BaseMessage
    list_display = ('text_view', 'html_view', 'level', 'session', 'created_time')
    search_fields = ('text',)
    ordering = ('session', 'created_time')

    def text_view(self, obj):
        if not obj.text: return
        abbr = escape(Truncator(obj.text).words(3))
        html = '<span title="{}">{}</span>'.format(escape(obj.text), abbr)
        return mark_safe(html)
    text_view.short_description = 'text'

    def html_view(self, obj):
        if not obj.html: return
        abbr = escape(Truncator(obj.html).words(3))
        html = '<span title="{}">{}</span>'.format(escape(obj.html), abbr)
        return mark_safe(html)
    html_view.short_description = 'html'



# -*- coding: utf-8 -*-

from django.contrib import admin
from django.utils.translation import gettext_lazy as _
from .models import MinkeSession


class StatusFilter(admin.SimpleListFilter):
    """
    Filter objects by the session-status of their current sessions.
    """
    title = _('Session-Status')
    parameter_name = 'minkestatus'

    def __init__(self, request, params, model, model_admin):
        super().__init__(request, params, model, model_admin)
        self.states = (s[0] for s in MinkeSession.SESSION_STATES)
        qs = model_admin.get_queryset(request)
        self.sessions = MinkeSession.objects.get_currents(request.user, qs)

    def has_output(self):
        return bool(self.sessions)

    def values(self):
        return self.value().split(',') if self.value() else list()

    def queryset(self, request, queryset):
        if not self.value(): return queryset

        ids = list()
        for status in self.values():
            for session in self.sessions.all():
                if not session.session_status == status: continue
                ids.append(session.minkeobj_id)
        return queryset.filter(id__in=ids)

    def choices(self, changelist):
        yield {
            'selected': not self.values(),
            'query_string': changelist.get_query_string({}, [self.parameter_name]),
            'display': _('All'),
        }
        for status in self.states:
            selected = status in self.values()
            if selected:
                values = set(self.values()) - set([status])
            else:
                values = set(self.values()) | set([status])
            if values:
                new_param = {self.parameter_name: ','.join(values)}
            else:
                new_param = dict()
            query_string = changelist.get_query_string(new_param, [self.parameter_name])
            yield {
                'selected': selected,
                'query_string': query_string,
                'display': status.title(),
            }

    def lookups(self, request, model_admin):
        return tuple()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-02-19 12:40
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Host',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('host', models.SlugField(max_length=128, unique=True)),
                ('user', models.SlugField(max_length=128)),
                ('hostname', models.CharField(max_length=128)),
                ('port', models.SmallIntegerField(blank=True, null=True)),
                ('hoststring', models.CharField(max_length=255, unique=True)),
                ('disabled', models.BooleanField(default=False)),
                ('locked', models.BooleanField(default=False)),
            ],
        ),
    ]



# -*- coding: utf-8 -*-
import sys

from django.core.management.base import BaseCommand
from django.core.management.base import CommandError

from ...models import Host
from ...models import MinkeSession


class Command(BaseCommand):
    help = 'Some cleanup-actions.'

    def add_arguments(self, parser):
        parser.add_argument(
            '-L', '--release-locks',
            action='store_true',
            help='Release locks on all hosts.')
        parser.add_argument(
            '-C', '--clear-current-sessions',
            action='store_true',
            help='Set current to False for all sessions.')
        parser.add_argument(
            '-S', '--clear-sessions',
            action='store_true',
            help='Delete all sessions.')

    def handle(self, *args, **options):
        if options['release_locks']:
            print(Host.objects.update(lock=None))
        if options['clear_current_sessions']:
            print(MinkeSession.objects.update(current=False))
        if options['clear_sessions']:
            print(MinkeSession.objects.all().delete())



# -*- coding: utf-8 -*-

import re
import datetime
from time import time
from collections import OrderedDict

from fabric2.runners import Result
from celery.task.control import revoke

from django.db import models
from django.db import transaction
from django.contrib.auth.models import User
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.fields import GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import FieldError
from django.utils.safestring import mark_safe
from django.template.loader import render_to_string
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _

from .exceptions import InvalidMinkeSetup
from .utils import JSONField


class MinkeSessionQuerySet(models.QuerySet):
    """
    Working with current sessions.
    Which are those that are rendered within the changelist.
    """
    def get_currents(self, user, minkeobjs):
        """
        Get all current sessions for a given user and minke-objects.
        """
        ct_query = ContentType.objects.filter(model=minkeobjs.model.__name__.lower())[0]
        qs = self.filter(minkeobj_type=ct_query, minkeobj_id__in=minkeobjs)
        return qs.filter(user=user, current=True)

    def clear_currents(self, user, minkeobjs):
        """
        Clear all current sessions for a given user and minke-objects.
        """
        return self.get_currents(user, minkeobjs).update(current=False)


# TODO: Add indexes for sessions, messages and commandresults!
class MinkeSession(models.Model):
    """
    The MinkeSession holds the data of any executed session and tracks its process.
    """
    objects = MinkeSessionQuerySet.as_manager()

    SESSION_STATES = (
        ('success', 0),
        ('warning', 1),
        ('error', 2))
    PROC_STATES = (
        ('initialized', 'waiting...'),
        ('running', 'running...'),
        ('completed', 'completed in {0:.1f} seconds'),
        ('stopping', 'stopping...'),
        ('stopped', 'stopped after {0:.1f} seconds'),
        ('canceled', 'canceled!'),
        ('failed', 'failed!'))
    SESSION_CHOICES = ((s[0], _(s[0])) for s in SESSION_STATES)
    PROC_CHOICES = ((s[0], _(s[0])) for s in PROC_STATES)

    class Meta:
        ordering = ('minkeobj_type_id', 'minkeobj_id', '-created_time')
        verbose_name = _('Session')
        verbose_name_plural = _('Sessions')

    # those fields will be derived from the session-class
    session_name = models.CharField(
        max_length=128,
        verbose_name=_('Session-name'),
        help_text=_('Class-name of the session-class.'))
    session_verbose_name = models.CharField(
        max_length=128,
        verbose_name=_("Session's verbose-name"),
        help_text=_('Verbose-name-attribute of the session-class.'))
    session_description = models.TextField(
        blank=True, null=True, max_length=128,
        verbose_name=_("Session's description"),
        help_text=_('Doc-string of the session-class.'))
    session_status = models.CharField(
        max_length=128, choices=SESSION_CHOICES,
        verbose_name=_("Session-status"),
        help_text=_('Mostly set by the session-code itself.'))
    session_data = JSONField(
        blank=True, null=True,
        verbose_name=_("Session's extra-data"),
        help_text=_('Data coming from a session-form.'))

    # the minkeobj to work on
    minkeobj_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    minkeobj_id = models.PositiveIntegerField()
    minkeobj = GenericForeignKey('minkeobj_type', 'minkeobj_id')

    # execution-data of the session
    user = models.ForeignKey(
        User, on_delete=models.CASCADE,
        verbose_name=_("User"),
        help_text=_('User that run this session.'))
    proc_status = models.CharField(
        max_length=128, choices=PROC_CHOICES,
        verbose_name=_("Process-status"),
        help_text=_('Status of session-processing.'))
    task_id = models.CharField(
        max_length=128, blank=True, null=True,
        verbose_name=_("Task-ID"),
        help_text=_('ID of the celery-task that run the session.'))
    start_time = models.DateTimeField(
        blank=True, null=True,
        verbose_name=_("Start-time"),
        help_text=_('Time the session has been started.'))
    end_time = models.DateTimeField(
        blank=True, null=True,
        verbose_name=_("End-time"),
        help_text=_('Time the session finished.'))
    run_time = models.DurationField(
        blank=True, null=True,
        verbose_name=_("Run-time"),
        help_text=_("Session's runtime."))
    created_time = models.DateTimeField(
        auto_now_add=True,
        verbose_name=_("Created-time"),
        help_text=_('Time the session has been initiated.'))
    current = models.BooleanField(default=True)

    def __str__(self):
        return '{} on {}'.format(self.session_name, self.minkeobj)

    def init(self, user, minkeobj, session_cls, session_data):
        """
        Initialize a session. Setup the session-attributes and save it.
        """
        self.proc_status = 'initialized'
        self.user = user
        self.minkeobj = minkeobj
        self.session_name = session_cls.__name__
        self.session_verbose_name = session_cls.verbose_name
        self.session_description = session_cls.__doc__
        self.session_data = session_data
        self.save()

    @transaction.atomic
    def start(self, task_id):
        """
        Start a session. Update proc_status, start_time and task_id.
        Since the cancel-method is called asynchrouniously to the whole session-
        processing, the start-, end- and cancel-method are each wrapped within a
        atomic transaction using select_for_update to protect them from interfering.
        """
        # We use the reloaded session for checks but update and save self.
        # On the database-level it doesn't make a difference from which object
        # we call the save-method.
        session = MinkeSession.objects.select_for_update().get(pk=self.id)
        if session.is_waiting:
            self.task_id = task_id
            self.proc_status = 'running'
            self.start_time = datetime.datetime.now()
            self.save(update_fields=['proc_status', 'start_time', 'task_id'])
            return True

    @transaction.atomic
    def cancel(self):
        """
        Cancel a session. Update proc_- and session_status.
        Since the cancel-method is called asynchrouniously to the whole session-
        processing, the start-, end- and cancel-method are each wrapped within a
        atomic transaction using select_for_update to protect them from interfering.
        """
        session = MinkeSession.objects.select_for_update().get(pk=self.id)
        if session.is_waiting:
            self.session_status = 'error'
            self.proc_status = 'canceled'
            self.save(update_fields=['proc_status', 'session_status'])
            return True
        elif session.proc_status == 'running':
            self.proc_status = 'stopping'
            self.save(update_fields=['proc_status'])
            revoke(session.task_id, signal='USR1', terminate=True)
            return True

    @transaction.atomic
    def end(self, failure=False):
        """
        End a session. Update proc_- and session_status, end_- and run_time.
        Since the cancel-method is called asynchrouniously to the whole session-
        processing, the start-, end- and cancel-method are each wrapped within a
        atomic transaction using select_for_update to protect them from interfering.
        """
        session = MinkeSession.objects.select_for_update().get(pk=self.id)
        if failure:
            self.session_status = 'error'
            self.proc_status = 'failed'
        elif session.proc_status == 'running':
            self.session_status = self.session_status or 'success'
            self.proc_status = 'completed'
        elif session.proc_status == 'stopping':
            self.session_status = 'error'
            self.proc_status = 'stopped'
        self.end_time = datetime.datetime.now()
        self.run_time = self.end_time - self.start_time
        fields = ['proc_status', 'session_status', 'end_time', 'run_time']
        self.save(update_fields=fields)

    @property
    def is_waiting(self):
        return self.proc_status == 'initialized'

    @property
    def is_running(self):
        return self.proc_status in ['running', 'stopping']

    @property
    def is_done(self):
        return self.proc_status in ['completed', 'canceled', 'stopped', 'failed']

    @property
    def proc_info(self):
        """
        Infos about the session-processing
        that will be rendered within the session-template.
        """
        info = next(s[1] for s in self.PROC_STATES if s[0] == self.proc_status)
        if self.run_time: return gettext(info).format(self.run_time.total_seconds())
        else: return gettext(info)

    def prnt(self):
        """
        Print a session and its messages.
        """
        width = 60
        pre_width = 7
        sep = ': '
        bg = dict(
            success = '\033[1;37;42m{}\033[0m'.format,
            warning = '\033[1;37;43m{}\033[0m'.format,
            error   = '\033[1;37;41m{}\033[0m'.format)
        fg = dict(
            info    = '\033[32m{}\033[39m'.format,
            warning = '\033[33m{}\033[39m'.format,
            error   = '\033[31m{}\033[39m'.format)
        ul = '\033[4m{}\033[0m'.format

        # print header
        minkeobj = str(self.minkeobj).ljust(width)
        status = self.session_status.upper().ljust(pre_width)
        print(bg[self.session_status](status + sep + minkeobj))

        # print messages
        msgs = list(self.messages.all())
        msg_count = len(msgs)
        for i, msg in enumerate(msgs, start=1):
            underlined = i < msg_count
            level = msg.level.ljust(pre_width)
            lines = msg.text.splitlines()
            for line in lines[:-1 if underlined else None]:
                print(fg[msg.level](level) + sep + line)
            if underlined:
                line = lines[-1].ljust(width)
                print(ul(fg[msg.level](level) + sep + line[:width]) + line[width:])


class CommandResult(Result, models.Model):
    """
    Add a db-layer to fabric's Result-class.
    """
    command = models.TextField(
        verbose_name=_('Command'),
        help_text=_('The command which was executed.'))
    exited = models.SmallIntegerField(
        verbose_name=_('Exit-status'),
        help_text=_('Exit-status returned by the command.'))
    stdout = models.TextField(
        blank=True, null=True,
        verbose_name=_('Stdout'),
        help_text=_('Standard-output of the command.'))
    stderr = models.TextField(
        blank=True, null=True,
        verbose_name=_('Stderr'),
        help_text=_('Standard-error of the command. '
                    '(unless the process was invoked via a pty, '
                    'in which case stderr and stdout are merged into stdout)'))
    shell = models.CharField(
        max_length=128,
        verbose_name=_('Shell'),
        help_text=_('The shell binary used for execution.'))
    encoding = models.CharField(
        max_length=128,
        verbose_name=_('Encoding'),
        help_text=_('The string encoding used by the local shell environment.'))
    pty = models.BooleanField(
        verbose_name=_('Pty'),
        help_text=_('A boolean describing whether the command was invoked with a pty or not'))
    created_time = models.DateTimeField(
        auto_now_add=True,
        verbose_name=_('Time of creation'),
        help_text=_('The datetime this command-result were added.'))
    session = models.ForeignKey(MinkeSession,
        on_delete=models.CASCADE,
        related_name='commands',
        verbose_name=_('Session'),
        help_text=_('Session whereas this command where executed.'))

    class Meta:
        ordering = ('session_id', 'created_time')
        verbose_name = _('Command-Result')
        verbose_name_plural = _('Command-Results')

    def __init__(self, *args, **kwargs):
        """
        This model could also be initiated as fabric's result-class.
        """
        try:
            # First we try to initiate the model.
            models.Model.__init__(self, *args, **kwargs)
        except TypeError:
            # If this fails, its a result-class-initiation.
            models.Model.__init__(self)
            Result.__init__(self, *args, **kwargs)

    def as_message(self):
        """
        Return this instance as an ExecutionMessage.
        """
        # FIXME: messages imports from models and vice versa.
        # We should find another solution here. Maybe define message-proxies
        # right here in the models-module?
        from .messages import ExecutionMessage
        return ExecutionMessage(self)


class BaseMessage(models.Model):
    """
    This is the database-layer for all messages. Proxies of this model are defined
    within the messages-module.
    """
    LEVELS = (
        ('info', 'info'),
        ('warning', 'warning'),
        ('error', 'error'))

    session = models.ForeignKey(MinkeSession,
        on_delete=models.CASCADE,
        related_name='messages',
        verbose_name=_('Session'),
        help_text=_('Session the message belongs to.'))
    level = models.CharField(
        max_length=128, choices=LEVELS,
        verbose_name=_('Message-level'),
        help_text=_('Level with which the message were added.'))
    text = models.TextField(
        verbose_name=_('Text'),
        help_text=_('Message-Text'))
    html = models.TextField(
        verbose_name=_('HTML'),
        help_text=_('Message as HTML'))
    created_time = models.DateTimeField(
        auto_now_add=True,
        verbose_name=_('Time of creation'),
        help_text=_('The datetime this message were added.'))

    class Meta:
        ordering = ('session_id', 'created_time')
        verbose_name = _('Message')
        verbose_name_plural = _('Messages')


class HostGroup(models.Model):
    """
    A Group of hosts. (Not sure if this is practical.)
    """
    name = models.CharField(
        max_length=255, unique=True,
        verbose_name=_('Group-Name'),
        help_text=_('Unique group-name.'))
    comment = models.TextField(
        blank=True, null=True,
        verbose_name=_('Comment'),
        help_text=_('Something about the group.'))
    config = models.CharField(
        max_length=255, blank=True, null=True,
        verbose_name=_('Fabric-/Invoke-config'),
        help_text=_('Use config as specified in MINKE_HOST_CONFIG.'))

    class Meta:
        ordering = ['name']
        verbose_name = _('Host-Group')
        verbose_name_plural = _('Host-Groups')

    def __str__(self):
        return self.name


class HostQuerySet(models.QuerySet):
    """
    Besides the get_lock-method this is an imitation of the minkemodel-queryset-api.
    """
    def get_lock(self):
        """
        Set a lock on all selected hosts.
        """
        # The most atomic way to get a lock is a update-query.
        # We use a timestamp to be able to identify the updated objects.
        timestamp = repr(time())
        self.filter(lock=None).update(lock=timestamp)
        return timestamp

    def get_hosts(self):
        """
        Return itself (minkemodel-api).
        """
        return self

    def host_filter(self, hosts):
        """
        Return an intersection of itself and the given hosts (minkemodel-api).
        """
        return self & hosts

    def select_related_hosts(self):
        """
        Return itself (minkemodel-api).
        """
        return self


class Host(models.Model):
    """
    This model is mainly a ssh-config.
    Each host represents an unique ssh-connection.
    It also imitates the minkemodel-api to normalize the way the engine
    runs sessions on them.
    """
    name = models.SlugField(
        max_length=128, unique=True,
        verbose_name=_('Name'),
        help_text=_('Unique name of the host. If Hostname is not specified fabric\'s '
                    'Connection-class will be initialized with this Name instead. '
                    'Specifying a Name only could be a sufficient Host-setup '
                    'if there is a valid ssh_config for it as lookup-pattern.'))
    verbose_name = models.CharField(
        max_length=255, blank=True, null=True,
        verbose_name=_('Verbose Name'),
        help_text=_('Verbose Host-Name.'))
    hostname = models.CharField(
        max_length=255, blank=True, null=True,
        verbose_name=_('HostName'),
        help_text=_('HostName could be either a ssh-config-lookup-pattern or a '
                    'real hostname to log into.'))
    username = models.CharField(
        max_length=255, blank=True, null=True,
        verbose_name=_('User'),
        help_text=_('User to login as.'))
    port = models.IntegerField(
        blank=True, null=True,
        verbose_name=_('Port number.'),
        help_text=_('Port number to connect on the remote host.'))
    comment = models.TextField(
        blank=True, null=True,
        verbose_name=_('Comment'),
        help_text=_('Something about the host.'))
    group = models.ForeignKey(HostGroup,
        blank=True, null=True, on_delete=models.SET_NULL,
        verbose_name=_('Hostgroup'),
        help_text=_('The group this host belongs to.'))
    config = models.CharField(
        max_length=255, blank=True, null=True,
        verbose_name=_('Fabric-/Invoke-config'),
        help_text=_('Use config as specified in MINKE_HOST_CONFIG.'))
    disabled = models.BooleanField(
        default=False,
        verbose_name=_('Disabled'),
        help_text=_('Disabled hosts won\'t be accessed by minke.'))
    lock = models.CharField(
        max_length=20, blank=True, null=True,
        verbose_name=_('Lock'),
        help_text=_('Locked hosts won\'t be accessed by minke.'
                    'To prevent intersection a host will be locked '
                    'while sessions are executed on it.'))

    objects = HostQuerySet.as_manager()
    sessions = GenericRelation(MinkeSession,
        content_type_field='minkeobj_type',
        object_id_field='minkeobj_id')

    def get_host(self):
        """
        Return itself (minkemodel-api).
        """
        return self

    def release_lock(self):
        """
        Release the host's lock.
        """
        self.lock = None
        self.save(update_fields=['lock'])

    class Meta:
        ordering = ['name']
        verbose_name = _('Host')
        verbose_name_plural = _('Hosts')

    def __str__(self):
        return self.name


class MinkeQuerySet(models.QuerySet):
    """
    A queryset-api to work with related hosts.
    This api is mainly used by the engine-module.
    """
    def get_hosts(self):
        """
        Get all hosts related to the objects of this queryset.
        """
        lookup = self.model.get_reverse_host_lookup() + '__in'
        try:
            return Host.objects.filter(**{lookup:self})
        except FieldError:
            msg = "Invalid reverse-host-lookup: {}".format(lookup)
            raise InvalidMinkeSetup(msg)

    def host_filter(self, hosts):
        """
        Get all objects related to the given hosts.
        """
        lookup = self.model.HOST_LOOKUP + '__in'
        try:
            return self.filter(**{lookup:hosts})
        except FieldError:
            msg = "Invalid host-lookup: {}".format(lookup)
            raise InvalidMinkeSetup(msg)

    def select_related_hosts(self):
        """
        Return a queryset which selects related hosts.
        """
        try:
            return self.select_related(self.model.HOST_LOOKUP)
        except FieldError:
            msg = "Invalid host-lookup: {}".format(self.model.HOST_LOOKUP)
            raise InvalidMinkeSetup(msg)


# TODO: implement a post-delete-signal-handler to cleanup sessions on deleting
# a minkeobject.
class MinkeModel(models.Model):
    """
    An abstract baseclass for all models on which sessions should be run.
    """
    objects = MinkeQuerySet.as_manager()
    sessions = GenericRelation(MinkeSession,
        content_type_field='minkeobj_type',
        object_id_field='minkeobj_id')

    HOST_LOOKUP = 'host'
    REVERSE_HOST_LOOKUP = None

    class Meta:
        abstract = True

    @classmethod
    def get_reverse_host_lookup(cls):
        """
        Derive a reverse lookup-term from HOST_LOOKUP.
        """
        if cls.REVERSE_HOST_LOOKUP:
            lookup = self.REVERSE_HOST_LOOKUP
        else:
            lookup_list = cls.HOST_LOOKUP.split('__')
            lookup_list.reverse()
            lookup_list.append(cls.__name__.lower())
            lookup = '__'.join(lookup_list[1:])
        return lookup

    def get_host(self):
        """
        Return the related host-instance.
        """
        host = self
        for attr in self.HOST_LOOKUP.split('__'):
            host = getattr(host, attr, None)
        if not isinstance(host, Host):
            msg = "Invalid host-lookup: {}".format(self.HOST_LOOKUP)
            raise InvalidMinkeSetup(msg)
        else:
            return host
